"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import pytest
import os
import importlib
import types
from inspect import isclass


def find_tests():
    testdir = os.path.realpath(os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        '..'
    ))
    res = []
    for root, dirs, files in os.walk(testdir):
        for name in files:
            p = os.path.join(root, name)
            if not p.endswith('.py') or not name.startswith('test'):
                continue
            res.extend(find_mod_disabled_tests(p, testdir))
    return res


def find_mod_disabled_tests(path, testdir):
    tests = []
    mname = 'biweeklybudget.tests.'
    mname += path.replace(testdir + '/', '').replace(
        '.py', '').replace('/', '.')
    spec = importlib.util.spec_from_file_location(mname, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    for x in dir(mod):
        if 'test' not in x.lower():
            continue
        a = getattr(mod, x)
        if isinstance(a, types.FunctionType) and 'donot' in x.lower():
            tests.append('%s.%s' % (mname, x))
        if isclass(a):
            if 'donot' in x.lower():
                tests.append('%s.%s' % (mname, x))
            for methname in dir(a):
                if 'test' not in methname.lower():
                    continue
                meth = getattr(a, methname)
                if not isinstance(meth, types.FunctionType):
                    continue
                if 'donot' in methname.lower():
                    tests.append('%s.%s.%s' % (mname, x, methname))
    return tests


@pytest.mark.skipif(sys.version_info[0:2] != (3, 6), reason='py36 only')
def test_disabled_tests():
    """
    Fail if any tests are disabled via my usual hackish "DONOT" prefix.
    """
    assert find_tests() == []



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import (
    Column, String, PrimaryKeyConstraint, ForeignKey
)
from sqlalchemy.orm import relationship

from biweeklybudget.models.base import Base, ModelAsDict

logger = logging.getLogger(__name__)


class PlaidAccount(Base, ModelAsDict):

    __tablename__ = 'plaid_accounts'
    __table_args__ = (
        PrimaryKeyConstraint('item_id', 'account_id'),
        {'mysql_engine': 'InnoDB'}
    )

    #: Plaid Item ID
    item_id = Column(
        String(70), ForeignKey('plaid_items.item_id'), nullable=False
    )

    #: Plaid Account ID
    account_id = Column(String(70), nullable=False)

    #: PlaidItem this PlaidAccount is associated with
    plaid_item = relationship('PlaidItem', uselist=False)

    #: Name of the account
    name = Column(String(100))

    #: mask
    mask = Column(String(20))

    #: Plaid account type
    account_type = Column(String(50))

    #: Plaid account subtype
    account_subtype = Column(String(50))

    def __repr__(self):
        return "<PlaidAccount(item_id=%s, account_id=%s)>" % (
            self.item_id, self.account_id
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import (
    Column, String, PrimaryKeyConstraint, Text, Numeric, Boolean, ForeignKey,
    Integer, inspect
)
from sqlalchemy.sql.expression import null
from sqlalchemy_utc import UtcDateTime
from sqlalchemy.orm import relationship
from pytz import UTC
from datetime import datetime
import logging
import re
from decimal import Decimal

from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.settings import RECONCILE_BEGIN_DATE

logger = logging.getLogger(__name__)


class OFXTransaction(Base, ModelAsDict):

    __tablename__ = 'ofx_trans'
    __table_args__ = (
        PrimaryKeyConstraint('account_id', 'fitid'),
        {'mysql_engine': 'InnoDB'}
    )

    #: Account ID this transaction is associated with
    account_id = Column(
        Integer, ForeignKey('accounts.id'), nullable=False
    )

    #: Account this transaction is associated with
    account = relationship('Account', uselist=False)

    #: OFXStatement ID this transaction was last seen in
    statement_id = Column(
        Integer, ForeignKey('ofx_statements.id'), nullable=False
    )

    #: OFXStatement this transaction was last seen in
    statement = relationship(
        "OFXStatement", backref="ofx_trans"
    )

    #: OFX - FITID
    fitid = Column(String(255))

    #: OFX - Transaction Type
    trans_type = Column(String(50))

    #: OFX - Date Posted
    date_posted = Column(UtcDateTime)

    #: OFX - Amount
    amount = Column(Numeric(precision=10, scale=4))  # Amount format from OFX

    #: OFX - Name
    name = Column(String(255), index=True)

    #: OFX - Memo
    memo = Column(String(255), index=True)

    #: OFX - SIC
    sic = Column(String(255))

    #: OFX - MCC
    mcc = Column(String(255))

    #: OFX - Checknum
    checknum = Column(String(32))

    # app-specific fields

    #: Description
    description = Column(String(254), index=True)

    #: Notes
    notes = Column(Text)

    #: Account's ``re_payment`` matched
    is_payment = Column(Boolean, default=False)

    #: Account's ``re_late_fee`` matched
    is_late_fee = Column(Boolean, default=False)

    #: Account's ``re_interest_charge`` matched
    is_interest_charge = Column(Boolean, default=False)

    #: Account's ``re_fee`` matched
    is_other_fee = Column(Boolean, default=False)

    #: Account's ``re_interest_paid`` matched
    is_interest_payment = Column(Boolean, default=False)

    #: The reconcile_id for the OFX Transaction
    reconcile_id = Column(Integer, ForeignKey('txn_reconciles.id'))

    def __repr__(self):
        return "<OFXTransaction(account_id='%s', fitid='%s')>" % (
            self.account_id, self.fitid
        )

    @staticmethod
    def params_from_ofxparser_transaction(t, acct_id, stmt, cat_memo=False):
        """
        Given an ofxparser.ofxparser.Transaction object, generate and return
        a dict of kwargs to create a new OFXTransaction.

        :param t: ofxparser transaction
        :type t: ``ofxparser.ofxparser.Transaction``
        :param acct_id: OFXAccount ID
        :type acct_id: int
        :param stmt: OFXStatement this transaction was on
        :type stmt: biweeklybudget.models.ofx_statement.OFXStatement
        :param cat_memo: whether or not to concatenate OFX Memo to Name
        :type cat_memo: bool
        :return: dict of kwargs to create an OFXTransaction
        :rtype: dict
        """
        if t.id is None:
            raise RuntimeError('Transaction has no ID: %s', vars(t))
        kwargs = {
            'account_id': acct_id,
            'statement': stmt,
            'memo': t.memo,
            'name': t.payee,
            'amount': t.amount,
            'trans_type': t.type,
            # Note that as of 0.16, OfxParser returns tz-naive UTC datetimes
            'date_posted': t.date.replace(tzinfo=UTC),
            'fitid': t.id,
            'sic': t.sic,
            'mcc': t.mcc
        }
        if cat_memo:
            del kwargs['memo']
            kwargs['name'] = t.payee + t.memo
        for x in ['mcc', 'sic', 'checknum']:
            if not hasattr(t, x):
                continue
            val = getattr(t, x)
            if val is not None and val != '':
                kwargs[x] = val
        return kwargs

    @property
    def account_amount(self):
        """
        Return the amount of the transaction, appropriately negated if the
        :py:class:`~.Account` for this transaction has
        :py:attr:`~.Account.negate_ofx_amounts` True.

        :return: amount, negated as appropriate
        :rtype: decimal.Decimal
        """
        if not self.account.negate_ofx_amounts:
            return self.amount
        return self.amount * Decimal(-1.0)

    @staticmethod
    def unreconciled(db):
        """
        Return a query to match all unreconciled OFXTransactions.

        :param db: active database session to use for queries
        :type db: sqlalchemy.orm.session.Session
        :return: query to match all unreconciled OFXTransactions
        :rtype: sqlalchemy.orm.query.Query
        """
        cutoff_date = datetime(
            RECONCILE_BEGIN_DATE.year, RECONCILE_BEGIN_DATE.month,
            RECONCILE_BEGIN_DATE.day, 0, 0, 0, tzinfo=UTC
        )
        return db.query(OFXTransaction).filter(
            OFXTransaction.reconcile.__eq__(null()),
            OFXTransaction.date_posted.__ge__(cutoff_date),
            OFXTransaction.account.has(reconcile_trans=True),
            OFXTransaction.is_payment.__ne__(True),
            OFXTransaction.is_late_fee.__ne__(True),
            OFXTransaction.is_interest_charge.__ne__(True),
            OFXTransaction.is_other_fee.__ne__(True),
            OFXTransaction.is_interest_payment.__ne__(True)
        )

    @property
    def first_statement_by_date(self):
        """
        Return the first OFXStatement on or after `self.date_posted`.

        :return: first OFXStatement on or after `self.date_posted`
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        sess = inspect(self).session
        res = sess.query(OFXStatement).filter(
            OFXStatement.account.__eq__(self.account),
            OFXStatement.as_of.__ge__(self.date_posted)
        ).order_by(OFXStatement.as_of.asc()).first()
        logger.debug('First statement for %s: %s',
                     self, res)
        return res

    def update_is_fields(self):
        """
        Method to update all ``is_*`` fields on this instance, given the
        ``re_*`` properties of :py:attr:`~.account`.
        """
        fields = {
            're_interest_charge': 'is_interest_charge',
            're_interest_paid': 'is_interest_payment',
            're_payment': 'is_payment',
            're_late_fee': 'is_late_fee',
            're_other_fee': 'is_other_fee'
        }
        acct = self.account
        if acct is None:
            from biweeklybudget.models.account import Account
            sess = inspect(self).session
            acct = sess.query(Account).get(self.account_id)
        for fname in fields.values():
            if (
                fname == 'is_interest_charge' and
                self.name == 'Interest Charged - MANUALLY ENTERED'
            ):
                continue
            setattr(self, fname, False)
        for acct_attr, self_attr in fields.items():
            if (
                self_attr == 'is_interest_charge' and
                self.name == 'Interest Charged - MANUALLY ENTERED'
            ):
                continue
            r_str = getattr(acct, acct_attr)
            if r_str is None:
                continue
            try:
                if re.match(r_str, self.name, re.I):
                    setattr(self, self_attr, True)
            except Exception:
                logger.error('Error performing regex comparison on %s using '
                             'Account %s field %s (%s)', self, acct,
                             acct_attr, r_str, exc_info=True)



from __future__ import absolute_import
from __future__ import unicode_literals
import argparse
import getpass
import io
import logging
import os
import os.path
import sys

from ofxhome import OFXHome

from biweeklybudget.vendored.ofxclient.account import BankAccount, BrokerageAccount, CreditCardAccount
from biweeklybudget.vendored.ofxclient.config import OfxConfig
from biweeklybudget.vendored.ofxclient.institution import Institution
from biweeklybudget.vendored.ofxclient.util import combined_download
from biweeklybudget.vendored.ofxclient.client import DEFAULT_OFX_VERSION

AUTO_OPEN_DOWNLOADS = 1
DOWNLOAD_DAYS = 30

GlobalConfig = None


def run():
    global GlobalConfig

    parser = argparse.ArgumentParser(prog='ofxclient')
    parser.add_argument('-a', '--account')
    parser.add_argument('-d', '--download', type=argparse.FileType('wb', 0))
    parser.add_argument('-o', '--open', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-c', '--config', help='config file path')
    parser.add_argument('--download-days', default=DOWNLOAD_DAYS, type=int, help='number of days to download (default: %s)' % DOWNLOAD_DAYS)
    parser.add_argument('--ofx-version', default=DEFAULT_OFX_VERSION, type=int, help='ofx version to use for new accounts (default: %s)' % DEFAULT_OFX_VERSION)
    args = parser.parse_args()

    if args.config:
        GlobalConfig = OfxConfig(file_name=args.config)
    else:
        GlobalConfig = OfxConfig()

    accounts = GlobalConfig.accounts()
    account_ids = [a.local_id() for a in accounts]

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    if args.download:
        if accounts:
            if args.account:
                a = GlobalConfig.account(args.account)
                ofxdata = a.download(days=args.download_days)
            else:
                ofxdata = combined_download(accounts, days=args.download_days)
            args.download.write(ofxdata.read())
            if args.open:
                open_with_ofx_handler(args.download.name)
            sys.exit(0)
        else:
            print("no accounts configured")

    main_menu(args)


def main_menu(args):
    while 1:
        menu_title("Main\nEdit %s to\nchange descriptions or ofx options" %
                   GlobalConfig.file_name)

        accounts = GlobalConfig.accounts()
        for idx, account in enumerate(accounts):
            menu_item(idx, account.long_description())

        menu_item('A', 'Add an account')
        if accounts:
            menu_item('D', 'Download all combined')

        menu_item('Q', 'Quit')

        choice = prompt().lower()
        if choice == 'a':
            add_account_menu(args)
        elif choice == 'd':
            if not accounts:
                print("no accounts on file")
            else:
                ofxdata = combined_download(accounts, days=args.download_days)
                wrote = write_and_handle_download(
                    ofxdata,
                    'combined_download.ofx'
                )
                print("wrote: %s" % wrote)
        elif choice in ['q', '']:
            return
        elif int(choice) < len(accounts):
            account = accounts[int(choice)]
            view_account_menu(account, args)


def add_account_menu(args):
    menu_title("Add account")
    while 1:
        print('------')
        print('Notice')
        print('------')
        print('You are about to search for bank connection information')
        print('on a third party website.  This means you are trusting')
        print('http://ofxhome.com and their security policies.')
        print('')
        print('You will be sending your bank name to this website.')
        print('------')
        query = prompt('bank name eg. "express" (enter to exit)> ')
        if query.lower() in ['']:
            return

        found = OFXHome.search(query)
        if not found:
            error("No banks found")
            continue

        while 1:
            for idx, bank in enumerate(found):
                menu_item(idx, bank['name'])
            choice = prompt().lower()
            if choice in ['q', '']:
                return
            elif int(choice) < len(found):
                bank = OFXHome.lookup(found[int(choice)]['id'])
                if login_check_menu(bank, args):
                    return


def view_account_menu(account, args):
    while 1:
        menu_title(account.long_description())

        institution = account.institution
        client = institution.client()

        print("Overview:")
        print("  Name:           %s" % account.description)
        print("  Account Number: %s" % account.number_masked())
        print("  Institution:    %s" % institution.description)
        print("  Main Type:      %s" % str(type(account)))
        if hasattr(account, 'routing_number'):
            print("  Routing Number: %s" % account.routing_number)
            print("  Sub Type:       %s" % account.account_type)
        if hasattr(account, 'broker_id'):
            print("  Broker ID:      %s" % account.broker_id)

        print("Nerdy Info:")
        print("  Download Up To:        %s days" % args.download_days)
        print("  Username:              %s" % institution.username)
        print("  Local Account ID:      %s" % account.local_id())
        print("  Local Institution ID:  %s" % institution.local_id())
        print("  FI Id:                 %s" % institution.id)
        print("  FI Org:                %s" % institution.org)
        print("  FI Url:                %s" % institution.url)
        if institution.broker_id:
            print("  FI Broker Id:          %s" % institution.broker_id)
        print("  Client Id:             %s" % client.id)
        print("  App Ver:               %s" % client.app_version)
        print("  App Id:                %s" % client.app_id)
        print("  OFX Ver:               %s" % client.ofx_version)
        print("  User-Agent header:     %s" % client.user_agent)
        print("  Accept header:         %s" % client.accept)
        print("  Config File:           %s" % GlobalConfig.file_name)

        menu_item('D', 'Download')
        choice = prompt().lower()
        if choice == 'd':
            out = account.download(days=args.download_days)
            wrote = write_and_handle_download(out,
                                              "%s.ofx" % account.local_id())
            print("wrote: %s" % wrote)
        return


def login_check_menu(bank_info, args):
    print('------')
    print('Notice')
    print('------')
    print('You are about to test to make sure your username and password')
    print('are correct.  This means you will be sending it to the URL below.')
    print('If the URL does not appear to belong to your bank then you should')
    print('exit this program by hitting CTRL-C.')
    print('  bank name: %s' % (bank_info['name']))
    print('  bank url:  %s' % (bank_info['url']))
    print('------')
    while 1:
        username = ''
        while not username:
            username = prompt('username> ')

        password = ''
        prompt_text = 'password> '
        if os.name == 'nt' and sys.version_info < (3, 0):
            prompt_text = prompt_text.encode('utf8')
        while not password:
            password = getpass.getpass(prompt=prompt_text)

        i = Institution(
            id=bank_info['fid'],
            org=bank_info['org'],
            url=bank_info['url'],
            broker_id=bank_info['brokerid'],
            description=bank_info['name'],
            username=username,
            password=password,
            client_args=client_args_for_bank(bank_info, args.ofx_version)
        )
        try:
            i.authenticate()
        except Exception as e:
            print("authentication failed: %s" % e)
            continue

        accounts = i.accounts()
        for a in accounts:
            GlobalConfig.add_account(a)
        GlobalConfig.save()
        return 1


def client_args_for_bank(bank_info, ofx_version):
    """
    Return the client arguments to use for a particular Institution, as found
    from ofxhome. This provides us with an extension point to override or
    augment ofxhome data for specific institutions, such as those that
    require specific User-Agent headers (or no User-Agent header).

    :param bank_info: OFXHome bank information for the institution, as returned
      by ``OFXHome.lookup()``
    :type bank_info: dict
    :param ofx_version: OFX Version argument specified on command line
    :type ofx_version: str
    :return: Client arguments for a specific institution
    :rtype: dict
    """
    client_args = {'ofx_version': str(ofx_version)}
    if 'ofx.discovercard.com' in bank_info['url']:
        # Discover needs no User-Agent and no Accept headers
        client_args['user_agent'] = False
        client_args['accept'] = False
    if 'www.accountonline.com' in bank_info['url']:
        # Citi needs no User-Agent header
        client_args['user_agent'] = False
    return client_args

def write_and_handle_download(ofx_data, name):
    outfile = io.open(name, 'w')
    outfile.write(ofx_data.read())
    outfile.close()
    if AUTO_OPEN_DOWNLOADS:
        open_with_ofx_handler(name)
    return os.path.abspath(name)


def prompt(text='choice> '):
    try:
        # python 2
        got = raw_input(text)
    except NameError:
        # python 3
        got = input(text)
    return got


def error(text=''):
    print("!! %s" % text)


def menu_item(key, description):
    print("(%s) %s" % (key, description))


def menu_title(name):
    print("+----------------------------------")
    print("%s" % name)
    print("+----------------------------------")


def open_with_ofx_handler(filename):
    import platform
    sysname = platform.system()
    if sysname == 'Darwin':
        os.system("/usr/bin/open '%s'" % filename)
    elif sysname == 'Windows':
        os.startfile(filename)
    else:
        # linux
        os.system("xdg-open '%s'" % filename)

if __name__ == '__main__':
    run()



"""Account add do_not_reconcile boolean

Revision ID: ed827c515edb
Revises: 5446559aecf7
Create Date: 2017-04-29 09:46:33.057376

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed827c515edb'
down_revision = '5446559aecf7'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'accounts',
        sa.Column(
            'reconcile_trans',
            sa.Boolean(),
            nullable=False,
            default=True
        )
    )


def downgrade():
    op.drop_column('accounts', 'reconcile_trans')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
import pytest
from datetime import date, datetime
from pytz import UTC
from decimal import Decimal

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.scheduled_transaction import ScheduledTransaction
from biweeklybudget.models.transaction import Transaction
from biweeklybudget.models.account import Account, AcctType
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.tests.conftest import get_db_engine
from biweeklybudget.tests.sqlhelpers import restore_mysqldump

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch
else:
    from unittest.mock import patch

pbm = 'biweeklybudget.biweeklypayperiod'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestSchedTransOrderingAndPeriodAssignment(AcceptanceHelper):

    def find_income_trans_id(self, db):
        return db.query(ScheduledTransaction).filter(
            ScheduledTransaction.description.__eq__('Income')
        ).one().id

    def test_0_clean_transactions(self, testdb):
        testdb.query(TxnReconcile).delete(synchronize_session='fetch')
        testdb.query(BudgetTransaction).delete(synchronize_session='fetch')
        testdb.query(Transaction).delete(synchronize_session='fetch')
        num_rows = testdb.query(
            ScheduledTransaction).delete(synchronize_session='fetch')
        assert num_rows == 6
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_1_confirm_pay_period_start(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        assert pp.start_date == date(2017, 4, 7)

    def test_2_add_data(self, testdb):
        acct = testdb.query(Account).get(1)
        budg = testdb.query(Budget).get(1)
        inc_budg = testdb.query(Budget).get(7)
        for daynum in range(1, 29):
            testdb.add(ScheduledTransaction(
                amount=Decimal('123.45'),
                description='ST_day_%d' % daynum,
                account=acct,
                budget=budg,
                day_of_month=daynum
            ))
        testdb.add(ScheduledTransaction(
            amount=Decimal('-1000.00'),
            description='Income',
            account=acct,
            budget=inc_budg,
            num_per_period=1
        ))
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_3_previous_pay_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 1), testdb
        )
        assert pp.start_date == date(2017, 3, 24)
        all_trans = pp.transactions_list
        all_monthly = [x for x in all_trans if x['sched_type'] == 'monthly']
        assert len(all_monthly) == 11
        # note monthly scheduled must have a day number <= 28
        assert all_monthly[0]['date'] == date(2017, 3, 24)
        assert all_monthly[1]['date'] == date(2017, 3, 25)
        assert all_monthly[2]['date'] == date(2017, 3, 26)
        assert all_monthly[3]['date'] == date(2017, 3, 27)
        assert all_monthly[4]['date'] == date(2017, 3, 28)
        assert all_monthly[5]['date'] == date(2017, 4, 1)
        assert all_monthly[6]['date'] == date(2017, 4, 2)
        assert all_monthly[7]['date'] == date(2017, 4, 3)
        assert all_monthly[8]['date'] == date(2017, 4, 4)
        assert all_monthly[9]['date'] == date(2017, 4, 5)
        assert all_monthly[10]['date'] == date(2017, 4, 6)

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_4_income_trans_is_first_previous_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 1), testdb
        )
        all_trans = pp.transactions_list
        income_id = self.find_income_trans_id(testdb)
        assert income_id is not None
        assert all_trans[0]['id'] == income_id

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_5_current_pay_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 13), testdb
        )
        assert pp.start_date == date(2017, 4, 7)
        all_trans = pp.transactions_list
        all_monthly = [x for x in all_trans if x['sched_type'] == 'monthly']
        assert len(all_monthly) == 14
        assert all_monthly[0]['date'] == date(2017, 4, 7)
        assert all_monthly[1]['date'] == date(2017, 4, 8)
        assert all_monthly[2]['date'] == date(2017, 4, 9)
        assert all_monthly[3]['date'] == date(2017, 4, 10)
        assert all_monthly[4]['date'] == date(2017, 4, 11)
        assert all_monthly[5]['date'] == date(2017, 4, 12)
        assert all_monthly[6]['date'] == date(2017, 4, 13)
        assert all_monthly[7]['date'] == date(2017, 4, 14)
        assert all_monthly[8]['date'] == date(2017, 4, 15)
        assert all_monthly[9]['date'] == date(2017, 4, 16)
        assert all_monthly[10]['date'] == date(2017, 4, 17)
        assert all_monthly[11]['date'] == date(2017, 4, 18)
        assert all_monthly[12]['date'] == date(2017, 4, 19)
        assert all_monthly[13]['date'] == date(2017, 4, 20)

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_6_income_trans_is_first_current_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 13), testdb
        )
        all_trans = pp.transactions_list
        income_id = self.find_income_trans_id(testdb)
        assert income_id is not None
        assert all_trans[0]['id'] == income_id

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_7_next_pay_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 5, 4), testdb
        )
        assert pp.start_date == date(2017, 4, 21)
        all_trans = pp.transactions_list
        all_monthly = [x for x in all_trans if x['sched_type'] == 'monthly']
        assert len(all_monthly) == 12
        # note monthly scheduled must have a day number <= 28
        assert all_monthly[0]['date'] == date(2017, 4, 21)
        assert all_monthly[1]['date'] == date(2017, 4, 22)
        assert all_monthly[2]['date'] == date(2017, 4, 23)
        assert all_monthly[3]['date'] == date(2017, 4, 24)
        assert all_monthly[4]['date'] == date(2017, 4, 25)
        assert all_monthly[5]['date'] == date(2017, 4, 26)
        assert all_monthly[6]['date'] == date(2017, 4, 27)
        assert all_monthly[7]['date'] == date(2017, 4, 28)
        assert all_monthly[8]['date'] == date(2017, 5, 1)
        assert all_monthly[9]['date'] == date(2017, 5, 2)
        assert all_monthly[10]['date'] == date(2017, 5, 3)
        assert all_monthly[11]['date'] == date(2017, 5, 4)

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_8_income_trans_is_first_next_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 5, 4), testdb
        )
        all_trans = pp.transactions_list
        income_id = self.find_income_trans_id(testdb)
        assert income_id is not None
        assert all_trans[0]['id'] == income_id


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestTransFromSchedTrans(AcceptanceHelper):

    def test_0_clean_transactions(self, testdb):
        testdb.query(TxnReconcile).delete(synchronize_session='fetch')
        testdb.query(BudgetTransaction).delete(synchronize_session='fetch')
        testdb.query(Transaction).delete(synchronize_session='fetch')
        num_rows = testdb.query(
            ScheduledTransaction).delete(synchronize_session='fetch')
        assert num_rows == 6
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_1_confirm_pay_period_start(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        assert pp.start_date == date(2017, 4, 7)

    def test_2_add_data(self, testdb):
        acct = testdb.query(Account).get(1)
        budg = testdb.query(Budget).get(1)
        budg2 = testdb.query(Budget).get(2)
        st_daynum = ScheduledTransaction(
            amount=Decimal('111.11'),
            description='ST_day_9',
            account=acct,
            budget=budg,
            day_of_month=9
        )
        testdb.add(st_daynum)
        t_daynum = Transaction(
            budget_amounts={budg: Decimal('111.33')},
            budgeted_amount=Decimal('111.11'),
            date=date(2017, 4, 9),
            description='Trans_ST_day_9',
            account=acct,
            planned_budget=budg,
            scheduled_trans=st_daynum,
        )
        testdb.add(t_daynum)
        testdb.add(TxnReconcile(
            note='foo',
            transaction=t_daynum
        ))
        st_pp1 = ScheduledTransaction(
            amount=Decimal('222.22'),
            description='ST_pp_1',
            account=acct,
            budget=budg,
            num_per_period=1
        )
        testdb.add(st_pp1)
        st_pp3 = ScheduledTransaction(
            amount=Decimal('333.33'),
            description='ST_pp_3',
            account=acct,
            budget=budg,
            num_per_period=3
        )
        testdb.add(st_pp3)
        t_pp3A = Transaction(
            budget_amounts={budg: Decimal('333.33')},
            budgeted_amount=Decimal('333.33'),
            date=date(2017, 4, 14),
            description='Trans_ST_pp_3_A',
            account=acct,
            planned_budget=budg,
            scheduled_trans=st_pp3,
        )
        testdb.add(t_pp3A)
        t_pp3B = Transaction(
            budget_amounts={budg: Decimal('333.33')},
            budgeted_amount=Decimal('333.33'),
            date=date(2017, 4, 15),
            description='Trans_ST_pp_3_B',
            account=acct,
            planned_budget=budg,
            scheduled_trans=st_pp3
        )
        testdb.add(t_pp3B)
        st_date = ScheduledTransaction(
            amount=Decimal('444.44'),
            description='ST_date',
            account=acct,
            budget=budg,
            date=date(2017, 4, 12)
        )
        testdb.add(st_date)
        t_date = Transaction(
            budget_amounts={budg: Decimal('444.44')},
            budgeted_amount=Decimal('444.44'),
            date=date(2017, 4, 12),
            description='Trans_ST_date',
            account=acct,
            planned_budget=budg,
            scheduled_trans=st_date
        )
        testdb.add(t_date)
        t_foo = Transaction(
            budget_amounts={budg: Decimal('555.55')},
            date=date(2017, 4, 8),
            description='Trans_foo',
            account=acct
        )
        testdb.add(t_foo)
        t_bar = Transaction(
            budget_amounts={
                budg: Decimal('666.66'),
                budg2: Decimal('100.00')
            },
            date=date(2017, 4, 16),
            description='Trans_bar',
            account=acct
        )
        testdb.add(t_bar)
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_3_ignore_scheduled_converted_to_real_trans(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 7), testdb
        )
        assert pp.start_date == date(2017, 4, 7)
        all_trans = pp.transactions_list
        assert all_trans == [
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('222.2200'),
                'budgeted_amount': None,
                'budgets': {
                    1: {'amount': Decimal('222.2200'), 'name': 'Periodic1'}
                },
                'date': None,
                'description': 'ST_pp_1',
                'id': 8,
                'reconcile_id': None,
                'sched_trans_id': None,
                'sched_type': 'per period',
                'type': 'ScheduledTransaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('333.3300'),
                'budgeted_amount': None,
                'budgets': {
                    1: {'amount': Decimal('333.3300'), 'name': 'Periodic1'}
                },
                'date': None,
                'description': 'ST_pp_3',
                'id': 9,
                'reconcile_id': None,
                'sched_trans_id': None,
                'sched_type': 'per period',
                'type': 'ScheduledTransaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('555.5500'),
                'budgeted_amount': None,
                'budgets': {
                    1: {'amount': Decimal('555.5500'), 'name': 'Periodic1'}
                },
                'date': date(2017, 4, 8),
                'description': 'Trans_foo',
                'id': 9,
                'planned_budget_id': None,
                'planned_budget_name': None,
                'reconcile_id': None,
                'sched_trans_id': None,
                'sched_type': None,
                'type': 'Transaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('111.3300'),
                'budgeted_amount': Decimal('111.1100'),
                'budgets': {
                    1: {'amount': Decimal('111.3300'), 'name': 'Periodic1'}
                },
                'date': date(2017, 4, 9),
                'description': 'Trans_ST_day_9',
                'id': 5,
                'planned_budget_id': 1,
                'planned_budget_name': 'Periodic1',
                'reconcile_id': 2,
                'sched_trans_id': 7,
                'sched_type': None,
                'type': 'Transaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('444.4400'),
                'budgeted_amount': Decimal('444.4400'),
                'budgets': {
                    1: {'amount': Decimal('444.4400'), 'name': 'Periodic1'}
                },
                'date': date(2017, 4, 12),
                'description': 'Trans_ST_date',
                'id': 8,
                'planned_budget_id': 1,
                'planned_budget_name': 'Periodic1',
                'reconcile_id': None,
                'sched_trans_id': 10,
                'sched_type': None,
                'type': 'Transaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('333.3300'),
                'budgeted_amount': Decimal('333.3300'),
                'budgets': {
                    1: {'amount': Decimal('333.3300'), 'name': 'Periodic1'}
                },
                'date': date(2017, 4, 14),
                'description': 'Trans_ST_pp_3_A',
                'id': 6,
                'planned_budget_id': 1,
                'planned_budget_name': 'Periodic1',
                'reconcile_id': None,
                'sched_trans_id': 9,
                'sched_type': None,
                'type': 'Transaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('333.3300'),
                'budgeted_amount': Decimal('333.3300'),
                'budgets': {
                    1: {'amount': Decimal('333.3300'), 'name': 'Periodic1'}
                },
                'date': date(2017, 4, 15),
                'description': 'Trans_ST_pp_3_B',
                'id': 7,
                'planned_budget_id': 1,
                'planned_budget_name': 'Periodic1',
                'reconcile_id': None,
                'sched_trans_id': 9,
                'sched_type': None,
                'type': 'Transaction'
            },
            {
                'account_id': 1,
                'account_name': 'BankOne',
                'amount': Decimal('766.6600'),
                'budgeted_amount': None,
                'budgets': {
                    1: {'amount': Decimal('666.6600'), 'name': 'Periodic1'},
                    2: {'amount': Decimal('100.0000'), 'name': 'Periodic2'}
                },
                'date': date(2017, 4, 16),
                'description': 'Trans_bar',
                'id': 10,
                'planned_budget_id': None,
                'planned_budget_name': None,
                'reconcile_id': None,
                'sched_trans_id': None,
                'sched_type': None,
                'type': 'Transaction'
            }
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestSums(AcceptanceHelper):

    def test_10_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, get_db_engine(), with_data=False)

    def test_11_add_account(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_12_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Standing',
            is_periodic=False,
            description='1Standing',
            current_balance=Decimal('987.65')
        ))
        testdb.add(Budget(
            name='2Income',
            is_periodic=True,
            description='2Income',
            starting_balance=Decimal('123.45'),
            is_income=True
        ))
        testdb.add(Budget(
            name='3Income',
            is_periodic=True,
            description='2Income',
            starting_balance=Decimal('0.0'),
            is_income=True
        ))
        testdb.add(Budget(
            name='4Periodic',
            is_periodic=True,
            description='4Periodic',
            starting_balance=Decimal('500.00')
        ))
        testdb.add(Budget(
            name='5Periodic',
            is_periodic=True,
            description='5Periodic',
            starting_balance=Decimal('100.00')
        ))
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_13_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        budgets = {x.id: x for x in testdb.query(Budget).all()}
        # Budget 3 Income Transaction
        t1 = Transaction(
            date=date(2017, 4, 7),
            budget_amounts={
                budgets[3]: Decimal('100.00'),
                budgets[4]: Decimal('50.00')
            },
            budgeted_amount=Decimal('100.00'),
            description='B3 Income',
            account=acct
        )
        testdb.add(t1)
        # Budget 3 Income ST
        testdb.add(ScheduledTransaction(
            amount=Decimal('99.00'),
            description='B3 Income ST',
            account=acct,
            budget=budgets[3],
            num_per_period=1
        ))
        # Budget 4 allocated greater than budgeted (500.00)
        testdb.add(ScheduledTransaction(
            amount=Decimal('250.00'),
            description='B4 ST',
            account=acct,
            budget=budgets[4],
            date=date(2017, 4, 10)
        ))
        t2 = Transaction(
            date=date(2017, 4, 11),
            budget_amounts={budgets[4]: Decimal('250.00')},
            description='B4 T no budgeted',
            account=acct
        )
        testdb.add(t2)
        t3 = Transaction(
            date=date(2017, 4, 12),
            budget_amounts={budgets[4]: Decimal('600.00')},
            budgeted_amount=Decimal('500.00'),
            description='B4 T budgeted',
            account=acct,
            planned_budget=budgets[4]
        )
        testdb.add(t3)
        # Budget 5 budgeted greater than allocated (100)
        testdb.add(ScheduledTransaction(
            amount=Decimal('2.00'),
            description='B5 ST',
            account=acct,
            budget=budgets[5],
            day_of_month=9
        ))
        t4 = Transaction(
            date=date(2017, 4, 13),
            description='B5 T',
            budget_amounts={budgets[5]: Decimal('3.00')},
            budgeted_amount=Decimal('1.00'),
            account=acct,
            planned_budget=budgets[5]
        )
        testdb.add(t4)
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_14_budget_sums(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        assert pp._data['budget_sums'] == {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('199.0'),
                'spent': Decimal('100.0'),
                'trans_total': Decimal('199.0'),
                'is_income': True,
                'remaining': Decimal('199.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('1000.0'),
                'spent': Decimal('900.0'),
                'trans_total': Decimal('1150.0'),
                'is_income': False,
                'remaining': Decimal('-650.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('3.0'),
                'spent': Decimal('3.0'),
                'trans_total': Decimal('5.0'),
                'is_income': False,
                'remaining': Decimal('95.0')
            }
        }

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_15_overall_sums(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        assert pp._data['overall_sums'] == {
            'allocated': Decimal('1100.0'),
            'spent': Decimal('903.0'),
            'income': Decimal('322.45'),
            'remaining': Decimal('-580.55')
        }

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_16_transaction_list_ordering(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        type_id = [[x['type'], x['id']] for x in pp.transactions_list]
        assert type_id == [
            ['ScheduledTransaction', 1],
            ['Transaction', 1],
            ['ScheduledTransaction', 3],
            ['ScheduledTransaction', 2],
            ['Transaction', 2],
            ['Transaction', 3],
            ['Transaction', 4]
        ]

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_17_spent_greater_than_allocated(self, testdb):
        acct = testdb.query(Account).get(1)
        budget = testdb.query(Budget).get(5)
        t = Transaction(
            date=date(2017, 4, 13),
            description='B6 T',
            budget_amounts={budget: Decimal('2032.0')},
            budgeted_amount=Decimal('32.0'),
            account=acct,
            planned_budget=budget
        )
        testdb.add(t)
        testdb.flush()
        testdb.commit()
        pp = BiweeklyPayPeriod.period_for_date(
            date(2017, 4, 10), testdb
        )
        assert pp._data['budget_sums'] == {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('199.0'),
                'spent': Decimal('100.0'),
                'trans_total': Decimal('199.0'),
                'is_income': True,
                'remaining': Decimal('199.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('1000.0'),
                'spent': Decimal('900.0'),
                'trans_total': Decimal('1150.0'),
                'is_income': False,
                'remaining': Decimal('-650.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('35.0'),
                'spent': Decimal('2035.0'),
                'trans_total': Decimal('2037.0'),
                'is_income': False,
                'remaining': Decimal('-1937.0')
            }
        }
        assert pp._data['overall_sums'] == {
            'allocated': Decimal('1100.0'),
            'spent': Decimal('2935.0'),
            'income': Decimal('322.45'),
            'remaining': Decimal('-2612.55')
        }

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_18_sums_for_empty_period(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2020, 4, 9), testdb
        )
        assert pp._data['budget_sums'] == {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('99.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('99.0'),
                'is_income': True,
                'remaining': Decimal('99.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('500.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('2.0'),
                'is_income': False,
                'remaining': Decimal('98.0')
            }
        }
        assert pp._data['overall_sums'] == {
            'allocated': Decimal('600.0'),
            'income': Decimal('222.45'),
            'remaining': Decimal('-377.55'),
            'spent': Decimal('0.0')
        }

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_80_issue201_setup(self, testdb):
        budg = Budget(
            name='6Periodic',
            is_periodic=True,
            description='6Periodic',
            starting_balance=Decimal('0.00')
        )
        testdb.add(budg)
        acct = testdb.query(Account).get(1)
        testdb.add(ScheduledTransaction(
            amount=Decimal('2.00'),
            description='B6 ST4',
            account=acct,
            budget=budg,
            day_of_month=11
        ))
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_81_issue201_validate_setup(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2020, 4, 9), testdb
        )
        assert pp._data['budget_sums'] == {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('99.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('99.0'),
                'is_income': True,
                'remaining': Decimal('99.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('500.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('2.0'),
                'is_income': False,
                'remaining': Decimal('98.0')
            },
            6: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('2.0'),
                'is_income': False,
                'remaining': Decimal('-2.0')
            }
        }
        assert pp._data['overall_sums'] == {
            'allocated': Decimal('602.0'),
            'income': Decimal('222.45'),
            'remaining': Decimal('-379.55'),
            'spent': Decimal('0.0')
        }

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_82_issue201_st_to_trans(self, testdb):
        budg = testdb.query(Budget).get(6)
        st = testdb.query(ScheduledTransaction).get(4)
        testdb.add(Transaction(
            date=date(2020, 4, 14),
            description='B6 T6 from ST4',
            budget_amounts={
                budg: Decimal('200.00')
            },
            budgeted_amount=Decimal('2.00'),
            account=testdb.query(Account).get(1),
            planned_budget=budg,
            scheduled_trans=st
        ))
        testdb.flush()
        testdb.commit()

    @patch('%s.settings.PAY_PERIOD_START_DATE' % pbm, date(2017, 4, 7))
    def test_83_issue201_confirm(self, testdb):
        pp = BiweeklyPayPeriod.period_for_date(
            date(2020, 4, 9), testdb
        )
        assert pp._data['budget_sums'] == {
            2: {
                'budget_amount': Decimal('123.45'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': True,
                'remaining': Decimal('123.45')
            },
            3: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('99.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('99.0'),
                'is_income': True,
                'remaining': Decimal('99.0')
            },
            4: {
                'budget_amount': Decimal('500.00'),
                'allocated': Decimal('0.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('0.0'),
                'is_income': False,
                'remaining': Decimal('500.0')
            },
            5: {
                'budget_amount': Decimal('100.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('0.0'),
                'trans_total': Decimal('2.0'),
                'is_income': False,
                'remaining': Decimal('98.0')
            },
            6: {
                'budget_amount': Decimal('0.0'),
                'allocated': Decimal('2.0'),
                'spent': Decimal('200.0'),
                'trans_total': Decimal('200.0'),
                'is_income': False,
                'remaining': Decimal('-200.0')
            }
        }
        assert pp._data['overall_sums'] == {
            'allocated': Decimal('602.0'),
            'income': Decimal('222.45'),
            'remaining': Decimal('-577.55'),
            'spent': Decimal('200.0')
        }



"""
add DB settings table

Revision ID: 32b34a664b5a
Revises: 9dc8545963be
Create Date: 2017-08-12 21:44:45.860272

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '32b34a664b5a'
down_revision = '9dc8545963be'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'settings',
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('value', sa.Text(), nullable=True),
        sa.Column('default_value', sa.Text(), nullable=True),
        sa.Column('is_json', sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint('name', name=op.f('pk_settings')),
        mysql_engine='InnoDB'
    )


def downgrade():
    op.drop_table('settings')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from biweeklybudget.utils import dtnow
from sqlalchemy import (
    Column, Integer, String, ForeignKey, ForeignKeyConstraint, UniqueConstraint,
    Index
)
from sqlalchemy_utc import UtcDateTime
from sqlalchemy.orm import relationship, backref
from biweeklybudget.models.base import Base, ModelAsDict


class TxnReconcile(Base, ModelAsDict):

    __tablename__ = 'txn_reconciles'
    # NOTE: __table_args__ is after column definitions

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Transaction ID
    txn_id = Column(Integer, ForeignKey('transactions.id'))

    #: Relationship - :py:class:`~.Transaction`
    transaction = relationship(
        "Transaction", backref=backref("reconcile", uselist=False),
        foreign_keys=[txn_id]
    )

    #: OFX Transaction FITID
    ofx_fitid = Column(String(255))

    #: OFX Transaction Account ID
    ofx_account_id = Column(Integer)

    #: Relationship - :py:class:`~.OFXTransaction`
    ofx_trans = relationship(
        "OFXTransaction", backref=backref("reconcile", uselist=False),
        foreign_keys=[ofx_fitid, ofx_account_id]
    )

    #: ReconcileRule ID; set if this reconcile was created by a rule
    rule_id = Column(Integer, ForeignKey('reconcile_rules.id'))

    #: Relationship - :py:class:`~.ReconcileRule` that created this reconcile,
    #: if any.
    rule = relationship("ReconcileRule")

    #: Notes
    note = Column(String(254))

    #: time when this reconcile was made
    reconciled_at = Column(UtcDateTime, default=dtnow())

    __table_args__ = (
        ForeignKeyConstraint(
            ['ofx_account_id', 'ofx_fitid'],
            ['ofx_trans.account_id', 'ofx_trans.fitid']
        ),
        UniqueConstraint('txn_id'),
        UniqueConstraint('ofx_account_id', 'ofx_fitid'),
        Index(
            'fk_txn_reconciles_txn_id_transactions',
            txn_id
        ),
        Index(
            'fk_txn_reconciles_ofx_account_id_ofx_trans',
            ofx_account_id,
            ofx_fitid
        ),
        {'mysql_engine': 'InnoDB'}
    )

    def __repr__(self):
        return "<TxnReconcile(id=%d)>" % (
            self.id
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
from sqlalchemy import create_engine, text


logger = logging.getLogger(__name__)


def uri_for_db(dbname):
    dbpass = ''
    if 'MYSQL_PASS' in os.environ:
        dbpass = ':%s' % os.environ['MYSQL_PASS']
    return '{scheme}://{user}{passwd}@{host}:{port}/{dbname}' \
           '?charset=utf8mb4'.format(
               scheme='mysql+pymysql',
               user=os.environ.get('MYSQL_USER', 'root'),
               passwd=dbpass,
               host=os.environ.get('MYSQL_HOST', '127.0.0.1'),
               port=os.environ.get('MYSQL_PORT', '3306'),
               dbname=dbname
           )


def empty_db_by_uri(uri):
    logger.info('Emptying database: %s', uri)
    engine = create_engine(uri)
    with engine.connect() as conn:
        tables = [r[0] for r in conn.execute(text('SHOW TABLES;'))]
        if len(tables) == 0:
            logger.info('Database already empty.')
            return
        logger.debug('Executing SQL: SET FOREIGN_KEY_CHECKS = 0')
        conn.execute(text('SET FOREIGN_KEY_CHECKS = 0'))
        sql = 'DROP TABLE %s;' % ', '.join(tables)
        logger.debug('Executing SQL: %s', sql)
        conn.execute(text(sql))
        logger.debug('Executing SQL: SET FOREIGN_KEY_CHECKS = 1')
        conn.execute(text('SET FOREIGN_KEY_CHECKS = 1'))
        conn.commit()
    engine.dispose()
    logger.info('DB emptied.')


def load_premigration_sql(uri):
    """
    Given a database URI, load the initial state SQL when Alembic was added
    to the project.
    """
    sqlpath = os.path.abspath(os.path.join(
        os.environ['TOXINIDIR'], 'biweeklybudget', 'tests', 'fixtures',
        'premigration_db_state.sql'
    ))
    logger.debug('Reading pre-migration DB state SQL from %s', sqlpath)
    sql = []
    with open(sqlpath, 'r') as fh:
        for line in fh.readlines():
            line = line.strip()
            if line != '':
                sql.append(line)
    engine = create_engine(uri)
    logger.info(
        'Loading SQL from pre-migration DB state into %s', engine.url.database
    )
    with engine.connect() as conn:
        for s in sql:
            logger.debug('Executing SQL: %s', s)
            conn.execute(text(s))
        conn.commit()
    engine.dispose()
    logger.info('Pre-migration DB state loaded into %s', engine.url.database)



#!/usr/bin/env python
"""
Script to build Docker image of the project and possibly test it.

Actions are largely driven by environment variables, and some information
about the version of biweeklybudget installed...

If ``DOCKER_TEST_TAG`` is set, just run tests against an existing (local) image
with that tag, and exit.

  - just build the image
  - test it unless ``TEST_DOCKER=="false"``
  - tag it appropriately
  - if checked out to a git tag according to versionfinder, or
    ``DOCKER_BUILD_VER`` is set, build Docker image for that version
    **from PyPI**. Otherwise, build using the sdist created by ``tox``.

The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import sys
import logging
import time
from git import Repo
import requests
import docker
from io import BytesIO
import tarfile
from biweeklybudget.version import VERSION
import subprocess
import argparse
from distutils.spawn import find_executable

FORMAT = "[%(asctime)s %(levelname)s] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)
logger = logging.getLogger()

for lname in ['versionfinder', 'pip', 'git', 'requests', 'docker']:
    _log = logging.getLogger(lname)
    _log.setLevel(logging.CRITICAL)
    _log.propagate = True

if sys.version_info[0:2] < (3, 6):
    raise SystemExit('ERROR: Docker build can only run under py >= 3.6')

DOCKER_IMG = 'python:3.14-alpine3.23'
PY_VERSION = '3.14'
ACCEPTANCE_ENV = 'acceptance'

DOCKERFILE_TEMPLATE = """
# biweeklybudget Dockerfile - http://github.com/jantman/biweeklybudget
FROM {docker_img}

ARG version
USER root

COPY requirements.txt /tmp/requirements.txt
COPY entrypoint.sh /tmp/entrypoint.sh
{copy}

RUN /usr/local/bin/pip install virtualenv
RUN /usr/local/bin/virtualenv --system-site-packages /app
RUN set -ex \
    && apk add --no-cache \
        fontconfig \
        libxml2 \
        libxml2-dev \
        libxslt \
        libxslt-dev \
        tini \
        git \
        py3-cryptography \
    && apk add --no-cache --virtual .build-deps \
        gcc \
        libffi-dev \
        linux-headers \
        make \
        musl-dev \
        openssl-dev \
    && /app/bin/pip install {install} \
    && /app/bin/pip install gunicorn==22.0.0 \
    && apk del .build-deps \
    && rm -Rf /root/.cache{versionfix}

RUN install -g root -o root -m 755 /tmp/entrypoint.sh /app/bin/entrypoint.sh

# default to using settings_example.py, and user can override as needed
ENV SETTINGS_MODULE=biweeklybudget.settings_example
ENV LANG=en_US.UTF-8

LABEL com.jasonantman.biweeklybudget.version=$version
LABEL maintainer "jason@jasonantman.com"
LABEL homepage "http://github.com/jantman/biweeklybudget"

EXPOSE 80
ENTRYPOINT ["/sbin/tini", "--"]
CMD ["/app/bin/entrypoint.sh"]
"""


class DockerImageBuilder(object):

    image_name = 'jantman/biweeklybudget'

    def __init__(self, toxinidir):
        """
        :param toxinidir: directory containing tox.ini
        :type toxinidir: str
        """
        self._toxinidir = toxinidir
        self._gitdir = os.path.join(self._toxinidir, '.git')
        logger.info(
            'Initializing DockerImageBuilder; toxinidir=%s gitdir=%s',
            self._toxinidir, self._gitdir
        )
        if not os.path.exists(self._gitdir) or not os.path.isdir(self._gitdir):
            raise RuntimeError(
                'Error: %s does not exist or is not a directory' % self._gitdir
            )
        logger.debug('Connecting to Docker')
        self._docker = docker.from_env()

    def _find_git_info(self):
        """
        Return information about the state of the Git repository tox is being
        run from.

        :return: dict with keys 'dirty' (bool), 'sha' (str), 'tag' (str or None)
        :rtype: dict
        """
        res = {}
        logger.debug('Checking git status...')
        repo = Repo(path=self._gitdir, search_parent_directories=False)
        res['sha'] = repo.head.commit.hexsha
        res['dirty'] = repo.is_dirty(untracked_files=True)
        if os.environ.get('GITHUB_ACTIONS') == 'true':
            res['dirty'] = False
        res['tag'] = None
        for tag in repo.tags:
            # each is a git.Tag object
            if tag.commit.hexsha == res['sha']:
                res['tag'] = tag.name
        logger.debug('Git info: %s', res)
        return res

    @property
    def build_ver(self):
        """
        Return the pypi version to build, or None if not a release build.

        :rtype: str
        :return: release build version or None
        """
        env_ver = os.environ.get('DOCKER_BUILD_VER', None)
        if env_ver is not None:
            logger.debug('build_ver %s based on DOCKER_BUILD_VER env var',
                         env_ver)
            return env_ver
        return None

    def _tag_for_local(self):
        """
        Return the Docker image tag for a local build

        :return: tag for local build
        :rtype: str
        """
        time_s = '%d' % int(time.time())
        if self.build_ver is not None:
            tag = '%s_%s' % (self.build_ver, time_s)
            logger.debug('Local tag (build_ver): %s', tag)
            return tag
        sha = self._gitinfo['sha'][:8]
        if self._gitinfo['dirty']:
            sha += '-dirty'
        tag = '%s_%s' % (sha, time_s)
        logger.debug('Local tag: %s', tag)
        return tag

    def build(self):
        self._gitinfo = self._find_git_info()
        tag = self._tag_for_local()
        logger.info('Local build; tag=%s', tag)
        self._check_tag(tag)
        img_tag = self._build_image(tag)
        if self._needs_test:
            self.test(img_tag)
        logger.info('Image "%s" built and tested.', img_tag)
        if os.environ.get('GITHUB_ACTIONS') == 'true':
            with open(os.environ['GITHUB_OUTPUT'], 'a') as fh:
                fh.write(f'DOCKER_IMG_TAG={img_tag}\n')
            return img_tag
        if self.build_ver is not None:
            print("To push release image to Docker Hub:")
            print('docker tag %s:%s %s:%s' % (
                self.image_name, img_tag, self.image_name, self.build_ver
            ))
            print('docker push %s:%s' % (self.image_name, self.build_ver))
            print('docker tag %s:%s %s:latest' % (
                self.image_name, img_tag, self.image_name
            ))
            print('docker push %s:latest' % self.image_name)
        return img_tag

    @property
    def _needs_test(self):
        """
        Return True if image needs to be tested, False otherwise.

        :return: whether image should be tested
        :rtype: bool
        """
        if os.environ.get('TEST_DOCKER', 'true') == 'false':
            return False
        return True

    def test(self, tag):
        """
        Test the image that was just built.

        :param tag: tag of built image
        :type tag: str
        """
        db_container = None
        container = None
        try:
            db_container = self._run_mysql()
            dbname = db_container.name
            kwargs = {
                'detach': True,
                'name': 'biweeklybudget-test-%s' % int(time.time()),
                'environment': {
                    'DB_CONNSTRING': 'mysql+pymysql://'
                        'root:root@mysql:3306/'
                        'budgetfoo?charset=utf8mb4',
                    # Use test settings module and frozen timestamp for
                    # acceptance tests to match expected time-based assertions
                    'SETTINGS_MODULE':
                        'biweeklybudget.tests.fixtures.test_settings',
                    'BIWEEKLYBUDGET_TEST_TIMESTAMP': '1501223084'
                },
                'links': {dbname: 'mysql'},
                'ports': {
                    '80/tcp': None
                }
            }
            img = '%s:%s' % (self.image_name, tag)
            logger.info('Docker run %s with kwargs: %s', img, kwargs)
            container = self._docker.containers.run(img, **kwargs)
            logger.info('Running biweeklybudget container; name=%s id=%s',
                        container.name, container.id)
            logger.info('Container status: %s', container.status)
            logger.info('Sleeping 20s for stabilization...')
            time.sleep(20)
            container.reload()
            logger.info('Container status: %s', container.status)
            if container.status != 'running':
                logger.critical('Container did not stay running! Logs:')
                logger.critical(
                    container.logs(stderr=True, stdout=True, stream=False,
                                   timestamps=True).decode()
                )
                raise RuntimeError('Container did not stay running')
            else:
                logger.info(
                    'Container logs:\n%s',
                    container.logs(stderr=True, stdout=True, stream=False,
                                   timestamps=True).decode()
                )
            # do the tests
            try:
                self._run_tests(db_container, container)
            except Exception as exc:
                logger.critical("Tests failed: %s", exc, exc_info=True)
                raise
        finally:
            try:
                logger.info(
                    'MariaDB container logs:\n%s',
                    db_container.logs(
                        stderr=True, stdout=True, stream=False, timestamps=True
                    ).decode()
                )
                logger.info(
                    'biweeklybudget container logs:\n%s',
                    container.logs(
                        stderr=True, stdout=True, stream=False, timestamps=True
                    ).decode()
                )
            except Exception:
                pass
            try:
                db_container.stop()
                db_container.remove()
            except Exception:
                pass
            try:
                container.stop()
                container.remove()
            except Exception:
                pass

    def _run_tests(self, db_container, container):
        """
        Run smoke tests against the container.

        :param db_container: MariaDB Docker container
        :type db_container: ``docker.models.containers.Container``
        :param container: biweeklybudget Docker container
        :type container: ``docker.models.containers.Container``
        """
        c_ip = self._container_ip(container)
        baseurl = 'http://%s' % c_ip
        failures = False
        r = requests.get('%s/' % baseurl)
        if r.status_code == 200:
            logger.info('GET /: OK (200)')
        else:
            logger.error('GET /: %s', r.status_code)
            failures = True
        exp = '<a class="navbar-brand" href="index.html">BiweeklyBudget</a>'
        if exp in r.text:
            logger.info('GET / content OK')
        else:
            logger.error('GET / content NOT OK: \n%s', r.text)
            failures = True
        r = requests.get('%s/payperiods' % baseurl)
        if r.status_code == 200:
            logger.info('GET /payperiods: OK (200)')
        else:
            logger.error('GET /payperiods: %s', r.status_code)
            failures = True
        try:
            self._test_scripts(container)
            logger.info('script tests PASSED')
        except Exception as exc:
            logger.error('script tests failed: %s', exc, exc_info=True)
            failures = True
        try:
            self._run_acceptance_tests(db_container, container)
            logger.info('Acceptance tests PASSED')
        except Exception as exc:
            logger.error('Acceptance tests FAILED: %s', exc, exc_info=True)
            failures = True
        if failures:
            raise RuntimeError('Tests FAILED.')

    def _run_acceptance_tests(self, db_container, container):
        """
        Run the ``acceptance38`` tox environment against the running container.

        :param db_container: MariaDB Docker container
        :type db_container: ``docker.models.containers.Container``
        :param container: biweeklybudget Docker container
        :type container: ``docker.models.containers.Container``
        :raises: RuntimeError
        """
        db_ip = self._container_ip(db_container)
        web_ip = self._container_ip(container)
        env = {
            'CI': os.environ.get('CI', 'true'),
            'LANG': os.environ.get('LANG'),
            'SETTINGS_MODULE': os.environ.get('SETTINGS_MODULE'),
            'VIRTUAL_ENV': os.environ.get('VIRTUAL_ENV'),
            'DB_CONNSTRING': 'mysql+pymysql://root:root@%s:3306/budgetfoo?'
                             'charset=utf8mb4' % db_ip,
            'BIWEEKLYBUDGET_TEST_BASE_URL': 'http://%s' % web_ip,
            'PATH': self._fixed_path
        }
        if os.path.exists(os.path.join(self._toxinidir, 'venv', 'bin', 'tox')):
            toxpath = os.path.join(self._toxinidir, 'venv', 'bin', 'tox')
        elif os.path.exists(os.path.join(self._toxinidir, 'bin', 'tox')):
            toxpath = os.path.join(self._toxinidir, 'bin', 'tox')
        else:
            toxpath = find_executable('tox')
        cmd = [toxpath, '-e', ACCEPTANCE_ENV]
        logger.info(
            'Running acceptance tests against container; args="%s" cwd=%s '
            'timeout=3000 env=%s', ' '.join(cmd), self._toxinidir, env
        )
        res = subprocess.run(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            cwd=self._toxinidir, timeout=3000, env=env
        )
        logger.info('Acceptance test tox process exited %d', res.returncode)
        if res.returncode != 0:
            logger.error(
                'ERROR: Acceptance test process failed with exitcode %d',
                res.returncode
            )
            logger.error(
                'Acceptance test output:\n%s', res.stdout.decode()
            )
            res.check_returncode()

    @property
    def _fixed_path(self):
        """
        Return the current PATH, fixed to remove the docker tox env.

        :return: sanitized path
        :rtype: str
        """
        res = []
        toxdir = os.path.join(self._toxinidir, '.tox')
        for p in os.environ['PATH'].split(':'):
            if not p.startswith(toxdir):
                res.append(p)
        return ':'.join(res)

    def _container_ip(self, container):
        """
        Get the local (bridge) IP address for the specified container.

        :param container: biweeklybudget Docker container
        :type container: ``docker.models.containers.Container``
        :return: Container's local (bridge) IP address
        :rtype: str
        """
        container.reload()
        cnet = container.attrs['NetworkSettings']
        # Docker API changed - IPAddress may be directly in NetworkSettings
        # (older versions) or inside Networks -> network_name (newer versions)
        c_ip = cnet.get('IPAddress')
        if not c_ip and 'Networks' in cnet:
            # Get IP from the first available network (usually 'bridge')
            for network_name, network_data in cnet['Networks'].items():
                c_ip = network_data.get('IPAddress')
                if c_ip:
                    logger.debug(
                        'Container IP %s from network %s', c_ip, network_name
                    )
                    break
        logger.debug('Container IP: %s / NetworkSettings: %s', c_ip, cnet)
        return c_ip

    def _test_scripts(self, container):
        """
        Test scripts/entrypoints in the container.

        :param container: biweeklybudget Docker container
        :type container: ``docker.models.containers.Container``
        """
        test_cmds = [
            {
                'cmd': '/app/bin/ofxclient --help',
                'output': '--ofx-version OFX_VERSION'
            },
            {
                'cmd': '/app/bin/wishlist2project -h',
                'output': 'Synchronize Amazon wishlists to projects'
            },
            {
                'cmd': '/app/bin/initdb --help',
                'output': 'Load initial data to DB'
            },
            {
                'cmd': '/app/bin/ofxbackfiller --help',
                'output': 'Backfill OFX from disk'
            },
            {
                'cmd': '/app/bin/ofxgetter --help',
                'output': 'Download OFX transactions'
            },
            {
                'cmd': '/app/bin/loaddata --help',
                'output': 'Load initial data to DB'
            }
        ]
        for cmd_info in test_cmds:
            logger.debug('Running: %s', cmd_info['cmd'])
            ecode, res = container.exec_run(cmd_info['cmd'])
            logger.debug('Command exited %d; output:\n%s', ecode, res)
            if 'output' in cmd_info and cmd_info['output'] not in res.decode():
                raise RuntimeError(
                    'Expected %s output to include "%s" but it did not' % (
                        cmd_info['cmd'], cmd_info['output']
                    )
                )
        logger.info('script tests SUCCEEDED')

    def _run_mysql(self):
        """
        Run a MySQL (well, MariaDB) container to test the Docker image
        against.

        :return: MySQL container object
        :rtype: docker.models.containers.Container
        """
        img = 'mariadb:10.4.7'
        kwargs = {
            'detach': True,
            'name': 'biweeklybudget-mariadb-%s' % int(time.time()),
            'environment': {
                'MYSQL_ROOT_PASSWORD': 'root',
                'MYSQL_ROOT_HOST': '%'
            },
            'ports': {
                '3306/tcp': None
            }
        }
        logger.debug('Running %s with kwargs: %s', img, kwargs)
        cont = self._docker.containers.run(img, **kwargs)
        logger.debug('MySQL container running; name=%s id=%s',
                     cont.name, cont.id)
        logger.info('Waiting 60s for MySQL container to stabilize...')
        time.sleep(60)
        count = 0
        while count < 20:
            count += 1
            logger.info('Creating database...')
            cmd = ('/usr/bin/mysql -uroot -proot -h 127.0.0.1 '
                   '-e "CREATE DATABASE budgetfoo;"')
            logger.debug('Running: %s', cmd)
            ecode, res = cont.exec_run(cmd)
            logger.debug('Command exited %d; output:\n%s', ecode, res)
            if 'ERROR' not in res.decode() and ecode == 0:
                logger.info('Database creation appears successful.')
                break
            logger.warning('Database creation errored; sleep 5s and retry')
            time.sleep(5)
        return cont

    def _build_image(self, tag):
        """
        Build Docker image, with the specified tag.

        :param tag: tag to assign to the image
        :type tag: str
        :return: tag assigned to the image
        :rtype: str
        """
        ctx = self._docker_context()
        kwargs = {
            'fileobj': ctx,
            'custom_context': True,
            'tag': '%s:%s' % (self.image_name, tag),
            'quiet': False,
            'nocache': True,
            'rm': True,
            'pull': True,
            'dockerfile': '/Dockerfile',
            'buildargs': {'version': tag},
            'decode': True
        }
        logger.info('Running docker build with args: %s', kwargs)
        logger.info('Build running; output:')
        error = None
        for line in self._docker.api.build(**kwargs):
            if 'errorDetail' in line:
                error = line['errorDetail']
            try:
                print(line['stream'])
            except Exception:
                print("\t%s" % line)
        if error is not None:
            raise RuntimeError(str(error))
        logger.info('Build complete for image: %s', tag)
        return tag

    def _check_tag(self, tag):
        """
        Confirm that the specified tag is not already present on Docker Hub or
        locally.

        :param tag: tag to check
        :type tag: str
        """
        self._check_tag_local(tag)
        logger.debug('Checking for tag on hub.docker.com')
        url = 'https://hub.docker.com/' \
              'v2/repositories/%s/tags/' % self.image_name
        res = requests.get(url)
        logger.debug('GET %s: %d', url, res.status_code)
        if res.status_code == 404:
            return
        tags = [r['name'] for r in res.json()['results']]
        logger.debug('hub.docker.com tags for %s: %s', self.image_name, tags)
        if tag in tags:
            raise RuntimeError(
                "ERROR: Tag '%s' already exists on hub.docker.com for "
                "image '%s'" % (tag, self.image_name)
            )

    def _check_tag_local(self, tag):
        """
        Confirm that the specified tag is not already present locally.

        :param tag: tag to check
        :type tag: str
        """
        expected_name = '%s:%s' % (self.image_name, tag)
        for img in self._docker.images.list():
            if expected_name in img.tags:
                raise RuntimeError(
                    'ERROR: Docker image "%s" already exists '
                    'locally' % expected_name
                )
        logger.debug('No "%s" docker image locally', expected_name)

    @property
    def _tox_dist_file(self):
        """
        Return the absolute path to the tox dist file.

        :return: path to tox dist file
        :rtype: str
        """
        return os.environ['TOX_PACKAGE']

    def _docker_context(self):
        """
        Return a BytesIO object containing a tarred Docker build context
        for the image.

        :return: Docker build context
        :rtype: io.BytesIO
        """
        logger.debug('Creating docker context (tarfile BytesIO)')
        b = BytesIO()
        tar = tarfile.open(fileobj=b, mode='w')
        self._tar_add_string_file(tar, 'Dockerfile', self._dockerfile())
        self._tar_add_string_file(tar, 'entrypoint.sh', self._entrypoint())
        tar.add(
            os.path.join(self._toxinidir, 'requirements.txt'),
            arcname='requirements.txt'
        )
        tar.add(self._tox_dist_file, arcname='biweeklybudget.tar.gz')
        tar.add(
            os.path.join(self._toxinidir, 'dev', 'tini_0.14.0.deb'),
            arcname='tini_0.14.0.deb'
        )
        tar.close()
        b.seek(0)
        logger.debug('Docker context created')
        return b

    def _string_to_tarfile(self, path, s):
        """
        Return a BytesIO object containing a tarfile containing one file at
        ``path``, with contents ``s``.

        :param path: path in the tar archive to write the file at
        :type path: str
        :param s: file contents
        :type s: str
        :return: tarfile object containing one file
        :rtype: io.BytesIO
        """
        b = BytesIO()
        tar = tarfile.open(fileobj=b, mode='w')
        self._tar_add_string_file(tar, path, s)
        tar.close()
        b.seek(0)
        return b

    def _tar_add_string_file(self, tarobj, fpath, content):
        """
        Given a tarfile object, add a file to it at ``fpath``, with content
        ``content``.

        Largely based on: http://stackoverflow.com/a/40392022

        :param tarobj: the tarfile to add to
        :type tarobj: tarfile.TarFile
        :param fpath: path to put the file at in the archive
        :type fpath: str
        :param content: file content
        :type content: str
        """
        logger.debug('Adding %d-length string to tarfile at %s',
                     len(content), fpath)
        data = content.encode('utf-8')
        f = BytesIO(data)
        info = tarfile.TarInfo(name=fpath)
        info.size = len(data)
        tarobj.addfile(tarinfo=info, fileobj=f)

    def _dockerfile(self):
        """
        Return the text of a Dockerfile to build the image from.

        :return: Dockerfile text
        :rtype: str
        """
        if self.build_ver is None:
            s_copy = 'COPY biweeklybudget.tar.gz /tmp/biweeklybudget.tar.gz'
            s_install = '/tmp/biweeklybudget.tar.gz'
            ver = self._gitinfo['sha'][:8]
            if self._gitinfo['dirty']:
                ver += '-dirty'
            s_versionfix = "&& /bin/sed -i " \
                           "\"s/^VERSION =.*/VERSION = '%s+git.%s'/\"" \
                           f" /app/lib/python{PY_VERSION}/site-packages/" \
                           f"biweeklybudget/version.py" % (
                               VERSION, ver
                           )
        else:
            s_copy = ''
            s_install = 'biweeklybudget==%s' % self.build_ver
            s_versionfix = ''
        s = DOCKERFILE_TEMPLATE.format(
            copy=s_copy,
            install=s_install,
            versionfix=s_versionfix,
            docker_img=DOCKER_IMG
        )
        logger.debug("Dockerfile:\n%s", s)
        return s

    def _entrypoint(self):
        """
        Generate the string contents of the entrypoint script.

        :return: entrypoint script contents
        :rtype: str
        """
        s = "#!/bin/sh -ex\n"
        s += "export PYTHONUNBUFFERED=true\n"
        s += "/app/bin/python /app/bin/initdb -vv \n"
        s += "/app/bin/gunicorn -w 4 -b :80 --log-file=- --access-logfile=- " \
             "--access-logformat " \
             "'%(h)s %(l)s %(u)s %(t)s \"%(r)s\" %(s)s %(b)s \"%(f)s\" " \
             "\"%(a)s\" [%(L)ss]' " \
             "biweeklybudget.flaskapp.app:app\n"
        logger.debug('Entrypoint script:\n%s', s)
        return s


def set_log_info():
    """set logger level to INFO"""
    set_log_level_format(logging.INFO,
                         '%(asctime)s %(levelname)s:%(name)s:%(message)s')


def set_log_debug():
    """set logger level to DEBUG, and debug-level output format"""
    set_log_level_format(
        logging.DEBUG,
        "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - "
        "%(name)s.%(funcName)s() ] %(message)s"
    )


def set_log_level_format(level, format):
    """
    Set logger level and format.

    :param level: logging level; see the :py:mod:`logging` constants.
    :type level: int
    :param format: logging formatter format string
    :type format: str
    """
    formatter = logging.Formatter(fmt=format)
    logger.handlers[0].setFormatter(formatter)
    logger.setLevel(level)


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument('TOXINIDIR', type=str)
    args = p.parse_args(sys.argv[1:])
    set_log_debug()
    toxinidir = args.TOXINIDIR
    b = DockerImageBuilder(toxinidir)
    test_tag = os.environ.get('DOCKER_TEST_TAG', None)
    if test_tag is not None:
        logger.info('TEST-ONLY RUN for tag: %s', test_tag)
        b.test(test_tag)
    else:
        b.build()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
import os

if os.environ.get('NEW_RELIC_LICENSE_KEY', '') != '':
    import newrelic.agent
    newrelic.agent.initialize()

# workaround for https://github.com/jantman/versionfinder/issues/5
# caused by versionfinder import in ``views/help.py``
try:
    import pip  # noqa
except (ImportError, KeyError):
    pass

from collections import namedtuple

import datatables
from flask import Flask

from biweeklybudget.db import init_db, cleanup_db
from biweeklybudget.flaskapp.jsonencoder import MagicJSONProvider
from biweeklybudget.utils import fix_werkzeug_logger

format = "%(asctime)s [%(levelname)s %(filename)s:%(lineno)s - " \
         "%(name)s.%(funcName)s() ] %(message)s"
logging.basicConfig(level=logging.DEBUG, format=format)
logger = logging.getLogger()

if 'BIWEEKLYBUDGET_LOG_FILE' in os.environ:
    # mainly for acceptance tests
    fhandler = logging.FileHandler(os.environ['BIWEEKLYBUDGET_LOG_FILE'])
    fhandler.setLevel(logging.DEBUG)
    fhandler.setFormatter(logging.Formatter(fmt=format))
    logger.addHandler(fhandler)

fix_werkzeug_logger()

app = Flask(__name__)
app.jinja_env.add_extension('jinja2.ext.loopcontrols')
# Flask 3.x uses json_provider_class instead of json_encoder
app.json_provider_class = MagicJSONProvider
app.json = MagicJSONProvider(app)
init_db()


def before_request():
    """
    When running in debug mode, clear jinja cache.
    """
    logger.warning('DEBUG MODE - Clearing jinja cache')
    app.jinja_env.cache = {}


@app.teardown_appcontext
def shutdown_session(exception=None):
    cleanup_db()


if app.debug:
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.before_request(before_request)
    app.jinja_env.auto_reload = True


# Monkey-patch datatables package for SQLAlchemy 2.0 compatibility
# The package uses query.join('relationship_name') which no longer works in
# SQLAlchemy 2.0 - it must pass the actual relationship attribute instead
_original_datatable_init = datatables.DataTable.__init__


def _patched_datatable_init(self, params, model, query, columns):
    self.params = params
    self.model = model
    self.query = query
    self.data = {}
    self.columns = []
    self.columns_dict = {}
    self.search_func = lambda qs, s: qs

    DataColumn = namedtuple("DataColumn", ("name", "model_name", "filter"))

    for col in columns:
        name, model_name, filter_func = None, None, None

        if isinstance(col, datatables.DataColumn):
            self.columns.append(col)
            continue
        elif isinstance(col, tuple):
            if len(col) == 3:
                name, model_name, filter_func = col
            elif len(col) == 2:
                if callable(col[1]):
                    name, filter_func = col
                    model_name = name
                else:
                    name, model_name = col
            else:
                raise ValueError("Columns must be a tuple of 2 to 3 elements")
        else:
            name, model_name = col, col

        d = DataColumn(name=name, model_name=model_name, filter=filter_func)
        self.columns.append(d)
        self.columns_dict[d.name] = d

    # SQLAlchemy 2.0 fix: use getattr to get the actual relationship attribute
    # instead of passing a string to query.join()
    for column in (col for col in self.columns if "." in col.model_name):
        relationship_name = column.model_name.split(".")[0]
        self.query = self.query.join(getattr(self.model, relationship_name))


datatables.DataTable.__init__ = _patched_datatable_init


from biweeklybudget.flaskapp.views import *  # noqa
from biweeklybudget.flaskapp.filters import *  # noqa
from biweeklybudget.flaskapp.jinja_tests import *  # noqa
from biweeklybudget.flaskapp.context_processors import *  # noqa
from biweeklybudget.flaskapp.cli_commands import *  # noqa



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from sqlalchemy import func
from decimal import Decimal
from time import sleep
from biweeklybudget.models.projects import Project, BoMItem
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from selenium.webdriver.common.by import By


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestProjects(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/projects')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Projects / Bill of Materials - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestProjectsView(AcceptanceHelper):

    def test_00_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is True
        b = testdb.query(Project).get(2)
        assert b is not None
        assert b.name == 'P2'
        assert b.notes == 'ProjectTwo'
        assert b.is_active is True
        b = testdb.query(Project).get(3)
        assert b is not None
        assert b.name == 'P3Inactive'
        assert b.notes == 'ProjectThreeInactive'
        assert b.is_active is False
        assert testdb.query(Project).with_entities(
            func.max(Project.id)
        ).scalar() == 3
        assert testdb.query(BoMItem).with_entities(
            func.max(BoMItem.id)
        ).scalar() == 5

    def test_01_table(self, base_url, selenium):
        self.get(selenium, base_url + '/projects')
        assert selenium.find_element(By.ID,
                                     'active-remaining-cost').get_attribute('innerHTML') == '$77.77'
        assert selenium.find_element(By.ID,
                                     'active-total-cost').get_attribute('innerHTML') == '$2,546.89'
        table = selenium.find_element(By.ID, 'table-projects')
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<a href="/projects/1">P1</a>',
                '$2,546.89',
                '$77.77',
                'yes <a onclick="deactivateProject(1);" href="#">'
                '(deactivate)</a>',
                'ProjectOne'
            ],
            [
                '<a href="/projects/2">P2</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(2);" href="#">'
                '(deactivate)</a>',
                'ProjectTwo'
            ],
            [
                '<a href="/projects/3">P3Inactive</a>',
                '$5.34',
                '$3.00',
                'NO <a onclick="activateProject(3);" href="#">'
                '(activate)</a>',
                'ProjectThreeInactive'
            ]
        ]

    def test_02_search(self, base_url, selenium):
        self.get(selenium, base_url + '/projects')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('Inact')
        sleep(1)
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-projects')
        )
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<a href="/projects/3">P3Inactive</a>',
                '$5.34',
                '$3.00',
                'NO <a onclick="activateProject(3);" href="#">(activate)</a>',
                'ProjectThreeInactive'
            ]
        ]

    def test_03_add(self, base_url, selenium):
        self.get(selenium, base_url + '/projects')
        name = selenium.find_element(By.ID, 'proj_frm_name')
        name.clear()
        name.send_keys('NewP')
        notes = selenium.find_element(By.ID, 'proj_frm_notes')
        notes.clear()
        notes.send_keys('My New Project')
        btn = selenium.find_element(By.ID, 'formSaveButton')
        btn.click()
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-projects')
        )
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<a href="/projects/4">NewP</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(4);" href="#">'
                '(deactivate)</a>',
                'My New Project'
            ],
            [
                '<a href="/projects/1">P1</a>',
                '$2,546.89',
                '$77.77',
                'yes <a onclick="deactivateProject(1);" href="#">'
                '(deactivate)</a>',
                'ProjectOne'
            ],
            [
                '<a href="/projects/2">P2</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(2);" href="#">'
                '(deactivate)</a>',
                'ProjectTwo'
            ],
            [
                '<a href="/projects/3">P3Inactive</a>',
                '$5.34',
                '$3.00',
                'NO <a onclick="activateProject(3);" href="#">'
                '(activate)</a>',
                'ProjectThreeInactive'
            ]
        ]

    def test_04_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is True
        b = testdb.query(Project).get(2)
        assert b is not None
        assert b.name == 'P2'
        assert b.notes == 'ProjectTwo'
        assert b.is_active is True
        b = testdb.query(Project).get(3)
        assert b is not None
        assert b.name == 'P3Inactive'
        assert b.notes == 'ProjectThreeInactive'
        assert b.is_active is False
        b = testdb.query(Project).get(4)
        assert b is not None
        assert b.name == 'NewP'
        assert b.notes == 'My New Project'
        assert b.is_active is True
        assert testdb.query(Project).with_entities(
            func.max(Project.id)
        ).scalar() == 4
        assert testdb.query(BoMItem).with_entities(
            func.max(BoMItem.id)
        ).scalar() == 5

    def test_05_deactivate(self, base_url, selenium):
        self.get(selenium, base_url + '/projects')
        link = selenium.find_element(By.XPATH,
                                     '//a[@onclick="deactivateProject(1);"]'
                                     )
        link.click()
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-projects')
        )
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<a href="/projects/4">NewP</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(4);" href="#">'
                '(deactivate)</a>',
                'My New Project'
            ],
            [
                '<a href="/projects/2">P2</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(2);" href="#">'
                '(deactivate)</a>',
                'ProjectTwo'
            ],
            [
                '<a href="/projects/1">P1</a>',
                '$2,546.89',
                '$77.77',
                'NO <a onclick="activateProject(1);" href="#">'
                '(activate)</a>',
                'ProjectOne'
            ],
            [
                '<a href="/projects/3">P3Inactive</a>',
                '$5.34',
                '$3.00',
                'NO <a onclick="activateProject(3);" href="#">'
                '(activate)</a>',
                'ProjectThreeInactive'
            ]
        ]

    def test_06_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is False
        b = testdb.query(Project).get(2)
        assert b is not None
        assert b.name == 'P2'
        assert b.notes == 'ProjectTwo'
        assert b.is_active is True
        b = testdb.query(Project).get(3)
        assert b is not None
        assert b.name == 'P3Inactive'
        assert b.notes == 'ProjectThreeInactive'
        assert b.is_active is False
        b = testdb.query(Project).get(4)
        assert b is not None
        assert b.name == 'NewP'
        assert b.notes == 'My New Project'
        assert b.is_active is True
        assert testdb.query(Project).with_entities(
            func.max(Project.id)
        ).scalar() == 4
        assert testdb.query(BoMItem).with_entities(
            func.max(BoMItem.id)
        ).scalar() == 5

    def test_07_activate(self, base_url, selenium):
        self.get(selenium, base_url + '/projects')
        link = selenium.find_element(By.XPATH,
                                     '//a[@onclick="activateProject(3);"]'
                                     )
        link.click()
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-projects')
        )
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<a href="/projects/4">NewP</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(4);" href="#">'
                '(deactivate)</a>',
                'My New Project'
            ],
            [
                '<a href="/projects/2">P2</a>',
                '$0.00',
                '$0.00',
                'yes <a onclick="deactivateProject(2);" href="#">'
                '(deactivate)</a>',
                'ProjectTwo'
            ],
            [
                '<a href="/projects/3">P3Inactive</a>',
                '$5.34',
                '$3.00',
                'yes <a onclick="deactivateProject(3);" href="#">'
                '(deactivate)</a>',
                'ProjectThreeInactive'
            ],
            [
                '<a href="/projects/1">P1</a>',
                '$2,546.89',
                '$77.77',
                'NO <a onclick="activateProject(1);" href="#">'
                '(activate)</a>',
                'ProjectOne'
            ]
        ]

    def test_08_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is False
        b = testdb.query(Project).get(2)
        assert b is not None
        assert b.name == 'P2'
        assert b.notes == 'ProjectTwo'
        assert b.is_active is True
        b = testdb.query(Project).get(3)
        assert b is not None
        assert b.name == 'P3Inactive'
        assert b.notes == 'ProjectThreeInactive'
        assert b.is_active is True
        b = testdb.query(Project).get(4)
        assert b is not None
        assert b.name == 'NewP'
        assert b.notes == 'My New Project'
        assert b.is_active is True
        assert testdb.query(Project).with_entities(
            func.max(Project.id)
        ).scalar() == 4
        assert testdb.query(BoMItem).with_entities(
            func.max(BoMItem.id)
        ).scalar() == 5


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestOneProjectView(AcceptanceHelper):

    def test_00_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is True
        assert len(
            testdb.query(BoMItem).filter(BoMItem.project_id.__eq__(1)).all()
        ) == 3
        i = testdb.query(BoMItem).get(1)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item1'
        assert i.notes == 'P1Item1Notes'
        assert i.unit_cost == Decimal('11.11')
        assert i.quantity == 1
        assert i.url is None
        assert i.is_active is True
        i = testdb.query(BoMItem).get(2)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item2'
        assert i.notes == 'P1Item2Notes'
        assert i.unit_cost == Decimal('22.22')
        assert i.quantity == 3
        assert i.url == 'http://item2.p1.com'
        assert i.is_active is True
        i = testdb.query(BoMItem).get(3)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item3'
        assert i.notes == 'P1Item3Notes'
        assert i.unit_cost == Decimal('1234.56')
        assert i.quantity == 2
        assert i.url == 'http://item3.p1.com'
        assert i.is_active is False

    def test_01_table(self, base_url, selenium):
        self.get(selenium, base_url + '/projects/1')
        # Top headings
        assert selenium.find_element(By.ID,
                                     'div-notes').get_attribute('innerHTML') == 'ProjectOne'
        assert selenium.find_element(By.ID,
                                     'div-remaining-amount').get_attribute('innerHTML') == '$77.77'
        assert selenium.find_element(By.ID,
                                     'div-total-amount').get_attribute('innerHTML') == '$2,546.89'
        # Table
        table = selenium.find_element(By.ID, 'table-items')
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<span style="float: left;">P1Item1</span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(1)">(edit)</a></span>',
                '1',
                '$11.11',
                '$11.11',
                'P1Item1Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item2.p1.com">P1Item2</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(2)">(edit)</a></span>',
                '3',
                '$22.22',
                '$66.66',
                'P1Item2Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item3.p1.com">P1Item3</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(3)">(edit)</a></span>',
                '2',
                '$1,234.56',
                '$2,469.12',
                'P1Item3Notes',
                '<span style="color: #a94442;">NO</span>'
            ],
        ]
        rows = self.tbody2trlist(table)
        assert 'inactive' not in rows[0].get_attribute('class')
        assert 'inactive' not in rows[1].get_attribute('class')
        assert 'inactive' in rows[2].get_attribute('class')

    def test_02_search(self, base_url, selenium):
        self.get(selenium, base_url + '/projects/1')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('Item2')
        sleep(1)
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-items')
        )
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<span style="float: left;">'
                '<a href="http://item2.p1.com">P1Item2</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(2)">(edit)</a></span>',
                '3',
                '$22.22',
                '$66.66',
                'P1Item2Notes',
                'yes'
            ]
        ]
        rows = self.tbody2trlist(table)
        assert 'inactive' not in rows[0].get_attribute('class')

    def test_03_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/projects/1')
        editlink = selenium.find_element(By.XPATH,
                                         '//a[@onclick="bomItemModal(3)"]'
                                         )
        modal, title, body = self.try_click_and_get_modal(selenium, editlink)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit BoM Item 3'
        id = body.find_element(By.ID, 'bom_frm_id')
        assert id.get_attribute('value') == '3'
        proj_id = body.find_element(By.ID, 'bom_frm_project_id')
        assert proj_id.get_attribute('value') == '1'
        name = body.find_element(By.ID, 'bom_frm_name')
        assert name.get_attribute('value') == 'P1Item3'
        notes = body.find_element(By.ID, 'bom_frm_notes')
        assert notes.get_attribute('value') == 'P1Item3Notes'
        quantity = body.find_element(By.ID, 'bom_frm_quantity')
        assert quantity.get_attribute('value') == '2'
        cost = body.find_element(By.ID, 'bom_frm_unit_cost')
        assert cost.get_attribute('value') == '1234.56'
        url = body.find_element(By.ID, 'bom_frm_url')
        assert url.get_attribute('value') == 'http://item3.p1.com'
        active = body.find_element(By.ID, 'bom_frm_active')
        assert active.is_selected() is False

    def test_04_edit_item(self, base_url, selenium):
        self.get(selenium, base_url + '/projects/1')
        editlink = selenium.find_element(By.XPATH,
                                         '//a[@onclick="bomItemModal(3)"]'
                                         )
        modal, title, body = self.try_click_and_get_modal(selenium, editlink)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit BoM Item 3'
        id = body.find_element(By.ID, 'bom_frm_id')
        assert id.get_attribute('value') == '3'
        proj_id = body.find_element(By.ID, 'bom_frm_project_id')
        assert proj_id.get_attribute('value') == '1'
        name = body.find_element(By.ID, 'bom_frm_name')
        name.send_keys('Edited')
        notes = body.find_element(By.ID, 'bom_frm_notes')
        notes.clear()
        notes.send_keys('Foo')
        quantity = body.find_element(By.ID, 'bom_frm_quantity')
        quantity.clear()
        quantity.send_keys('3')
        cost = body.find_element(By.ID, 'bom_frm_unit_cost')
        cost.clear()
        cost.send_keys('2.22')
        url = body.find_element(By.ID, 'bom_frm_url')
        assert url.get_attribute('value') == 'http://item3.p1.com'
        url.send_keys('/edited')
        active = body.find_element(By.ID, 'bom_frm_active')
        assert active.is_selected() is False
        active.click()
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved BoMItem 3 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # Table
        table = selenium.find_element(By.ID, 'table-items')
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<span style="float: left;">P1Item1</span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(1)">(edit)</a></span>',
                '1',
                '$11.11',
                '$11.11',
                'P1Item1Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item2.p1.com">P1Item2</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(2)">(edit)</a></span>',
                '3',
                '$22.22',
                '$66.66',
                'P1Item2Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item3.p1.com/edited">P1Item3Edited</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(3)">(edit)</a></span>',
                '3',
                '$2.22',
                '$6.66',
                'Foo',
                'yes'
            ],
        ]
        rows = self.tbody2trlist(table)
        assert 'inactive' not in rows[0].get_attribute('class')
        assert 'inactive' not in rows[1].get_attribute('class')
        assert 'inactive' not in rows[2].get_attribute('class')
        self.wait_for_jquery_done(selenium)
        assert selenium.find_element(By.ID,
                                     'div-notes').get_attribute('innerHTML') == 'ProjectOne'
        assert selenium.find_element(By.ID,
                                     'div-remaining-amount').get_attribute('innerHTML') == '$84.43'
        assert selenium.find_element(By.ID,
                                     'div-total-amount').get_attribute('innerHTML') == '$84.43'

    def test_05_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is True
        assert len(
            testdb.query(BoMItem).filter(BoMItem.project_id.__eq__(1)).all()
        ) == 3
        i = testdb.query(BoMItem).get(1)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item1'
        assert i.notes == 'P1Item1Notes'
        assert i.unit_cost == Decimal('11.11')
        assert i.quantity == 1
        assert i.url is None
        assert i.is_active is True
        i = testdb.query(BoMItem).get(2)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item2'
        assert i.notes == 'P1Item2Notes'
        assert i.unit_cost == Decimal('22.22')
        assert i.quantity == 3
        assert i.url == 'http://item2.p1.com'
        assert i.is_active is True
        i = testdb.query(BoMItem).get(3)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item3Edited'
        assert i.notes == 'Foo'
        assert i.unit_cost == Decimal('2.22')
        assert i.quantity == 3
        assert i.url == 'http://item3.p1.com/edited'
        assert i.is_active is True

    def test_06_add_item(self, base_url, selenium):
        self.get(selenium, base_url + '/projects/1')
        editlink = selenium.find_element(By.ID, 'btn_add_item')
        modal, title, body = self.try_click_and_get_modal(selenium, editlink)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New BoM Item'
        id = body.find_element(By.ID, 'bom_frm_id')
        assert id.get_attribute('value') == ''
        proj_id = body.find_element(By.ID, 'bom_frm_project_id')
        assert proj_id.get_attribute('value') == '1'
        name = body.find_element(By.ID, 'bom_frm_name')
        name.send_keys('NewItem4')
        notes = body.find_element(By.ID, 'bom_frm_notes')
        notes.clear()
        notes.send_keys('FourNotes')
        quantity = body.find_element(By.ID, 'bom_frm_quantity')
        quantity.clear()
        quantity.send_keys('5')
        cost = body.find_element(By.ID, 'bom_frm_unit_cost')
        cost.clear()
        cost.send_keys('12.34')
        url = body.find_element(By.ID, 'bom_frm_url')
        url.send_keys('http://item4.com')
        active = body.find_element(By.ID, 'bom_frm_active')
        assert active.is_selected()
        active.click()
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved BoMItem 6 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # Table
        table = selenium.find_element(By.ID, 'table-items')
        htmls = self.inner_htmls(self.tbody2elemlist(table))
        assert htmls == [
            [
                '<span style="float: left;">P1Item1</span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(1)">(edit)</a></span>',
                '1',
                '$11.11',
                '$11.11',
                'P1Item1Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item2.p1.com">P1Item2</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(2)">(edit)</a></span>',
                '3',
                '$22.22',
                '$66.66',
                'P1Item2Notes',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item3.p1.com/edited">P1Item3Edited</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(3)">(edit)</a></span>',
                '3',
                '$2.22',
                '$6.66',
                'Foo',
                'yes'
            ],
            [
                '<span style="float: left;">'
                '<a href="http://item4.com">NewItem4</a></span>'
                '<span style="float: right;">'
                '<a href="#" onclick="bomItemModal(6)">(edit)</a></span>',
                '5',
                '$12.34',
                '$61.70',
                'FourNotes',
                '<span style="color: #a94442;">NO</span>'
            ]
        ]
        rows = self.tbody2trlist(table)
        assert 'inactive' not in rows[0].get_attribute('class')
        assert 'inactive' not in rows[1].get_attribute('class')
        assert 'inactive' not in rows[2].get_attribute('class')
        assert selenium.find_element(By.ID,
                                     'div-notes').get_attribute('innerHTML') == 'ProjectOne'
        assert selenium.find_element(By.ID,
                                     'div-remaining-amount').get_attribute('innerHTML') == '$84.43'
        assert selenium.find_element(By.ID,
                                     'div-total-amount').get_attribute('innerHTML') == '$146.13'

    def test_07_verify_db(self, testdb):
        b = testdb.query(Project).get(1)
        assert b is not None
        assert b.name == 'P1'
        assert b.notes == 'ProjectOne'
        assert b.is_active is True
        assert len(
            testdb.query(BoMItem).filter(BoMItem.project_id.__eq__(1)).all()
        ) == 4
        i = testdb.query(BoMItem).get(1)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item1'
        assert i.notes == 'P1Item1Notes'
        assert i.unit_cost == Decimal('11.11')
        assert i.quantity == 1
        assert i.url is None
        assert i.is_active is True
        i = testdb.query(BoMItem).get(2)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item2'
        assert i.notes == 'P1Item2Notes'
        assert i.unit_cost == Decimal('22.22')
        assert i.quantity == 3
        assert i.url == 'http://item2.p1.com'
        assert i.is_active is True
        i = testdb.query(BoMItem).get(3)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'P1Item3Edited'
        assert i.notes == 'Foo'
        assert i.unit_cost == Decimal('2.22')
        assert i.quantity == 3
        assert i.url == 'http://item3.p1.com/edited'
        assert i.is_active is True
        i = testdb.query(BoMItem).get(6)
        assert i is not None
        assert i.project_id == 1
        assert i.name == 'NewItem4'
        assert i.notes == 'FourNotes'
        assert i.unit_cost == Decimal('12.34')
        assert i.quantity == 5
        assert i.url == 'http://item4.com'
        assert i.is_active is False



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import jsonify, request
from datetime import datetime
from decimal import Decimal

logger = logging.getLogger(__name__)


class FormHandlerView(MethodView):

    def post(self):
        """
        Handle a POST request for a form. Validate it, if valid update the DB.

        Returns a JSON hash with the following structure:

        'errors' -> hash of field names to list of error strings
        'error_message' -> string error message
        'success' -> boolean
        'success_message' -> string success message
        """
        data = request.get_json(silent=True)
        if data is None:
            # not JSON, must be form encoded
            data = request.form.to_dict()
        res = self.validate(data)
        if res is not None:
            logger.info('Form validation failed. data=%s errors=%s',
                        data, res)
            return jsonify({
                'success': False,
                'errors': res
            })
        try:
            res = self.submit(data)
        except Exception as ex:
            logger.warning('Form submission failed. data=%s', data,
                           exc_info=True)
            return jsonify({
                'success': False,
                'error_message': str(ex)
            })
        if isinstance(res, type({})):
            return jsonify(res)
        return jsonify({
            'success': True,
            'success_message': res
        })

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        raise NotImplementedError()

    def _validate_int(self, key, data, errors):
        """
        Validate an integer field.

        :param key: the key in data to look at
        :type key: str
        :param data: the form data
        :type data: dict
        :param err_list: list of error messages for the field
        :type err_list: dict
        :return: updated err_list
        :rtype: dict
        """
        try:
            x = int(data[key])
            assert data[key] == '%d' % x
        except Exception:
            errors[key].append('Invalid integer value: "%s"' % data[key])
        return errors

    def _validate_float(self, key, data, errors):
        """
        Validate a float field.

        :param key: the key in data to look at
        :type key: str
        :param data: the form data
        :type data: dict
        :param err_list: list of error messages for the field
        :type err_list: dict
        :return: updated err_list
        :rtype: dict
        """
        try:
            x = float(data[key])
            assert data[key].startswith('%s' % x)
        except Exception:
            errors[key].append('Invalid float value: "%s"' % data[key])
        return errors

    def _validate_decimal(self, key, data, errors):
        """
        Validate a Decimal field.

        :param key: the key in data to look at
        :type key: str
        :param data: the form data
        :type data: dict
        :param err_list: list of error messages for the field
        :type err_list: dict
        :return: updated err_list
        :rtype: dict
        """
        try:
            Decimal(data[key])
        except Exception:
            errors[key].append('Invalid Decimal value: "%s"' % data[key])
        return errors

    def _validate_date_ymd(self, key, data, errors):
        """
        Validate a YYYY-mm-dd date field.

        :param key: the key in data to look at
        :type key: str
        :param data: the form data
        :type data: dict
        :param err_list: list of error messages for the field
        :type err_list: dict
        :return: updated err_list
        :rtype: dict
        """
        if data[key].strip() == '':
            errors[key].append('Date cannot be empty')
            return errors
        try:
            datetime.strptime(data[key], '%Y-%m-%d').date()
        except Exception:
            errors[key].append(
                'Date "%s" is not valid (YYYY-MM-DD)' % data[key]
            )
        return errors

    def _validate_not_empty(self, key, data, errors):
        """
        Validate that a string is not empty.

        :param key: the key in data to look at
        :type key: str
        :param data: the form data
        :type data: dict
        :param err_list: list of error messages for the field
        :type err_list: dict
        :return: updated err_list
        :rtype: dict
        """
        if data[key].strip() == '':
            errors[key].append('Cannot be empty')
        return errors

    def fix_string(self, s):
        """
        Strip a string. If the result is empty, return None. Otherwise return
        the result.

        :param s: form data value
        :type s: str
        :return: stripped string or None
        """
        s = s.strip()
        if s == '':
            return None
        return s

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB
        :rtype: str
        """
        raise NotImplementedError()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import time
from werkzeug.serving import run_simple, WSGIRequestHandler
from flask.cli import pass_script_info

import click


class CustomLoggingWSGIRequestHandler(WSGIRequestHandler):
    """Extend werkzeug request handler to include processing time in logs"""

    def handle(self):
        self._time_started = time.time()
        rv = super(CustomLoggingWSGIRequestHandler, self).handle()
        return rv

    def send_response(self, *args, **kw):
        self._time_finished = time.time()
        super(CustomLoggingWSGIRequestHandler, self).send_response(*args, **kw)

    def log_request(self, code='-', size='-'):
        duration = int((self._time_finished - self._time_started) * 1000)
        self.log(
            'info', '"%s" %s %s [%sms]',
            self.requestline, code, size, duration
        )


def template_paths():
    """
    Return a list of all Flask app template paths, to auto-reload on change.

    from http://stackoverflow.com/a/41666467/211734

    :return: list of all template paths
    :rtype: list
    """
    extra_dirs = [
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'templates'
        ),
    ]

    extra_files = extra_dirs[:]
    for extra_dir in extra_dirs:
        for dirname, dirs, files in os.walk(extra_dir):
            for filename in files:
                filename = os.path.join(dirname, filename)
                if os.path.isfile(filename):
                    extra_files.append(filename)
    return extra_files


@click.command('rundev', short_help='Runs a development server with reloading')
@click.option('--host', '-h', default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p', default=5000,
              help='The port to bind to.')
@click.option('--with-threads/--without-threads', default=False,
              help='Enable or disable multithreading.')
@pass_script_info
def rundev_command(info, host, port, with_threads):
    """
    Modified from the upstream ``flask.cli.run_command`` to add some debugging
    and reloading features
    """
    # Set a global flag that indicates that we were invoked from the
    # command line interface provided server command.  This is detected
    # by Flask.run to make the call into a no-op.  This is necessary to
    # avoid ugly errors when the script that is loaded here also attempts
    # to start a server.
    os.environ['FLASK_RUN_FROM_CLI_SERVER'] = '1'

    os.environ['FLASK_DEBUG'] = '1'

    debug = True
    reload = True
    debugger = True

    # Load the Flask app directly (DispatchingApp was removed in Flask 2.3)
    app = info.load_app()

    # Extra startup messages.  This depends a bit on Werkzeug internals to
    # not double execute when the reloader kicks in.
    if os.environ.get('WERKZEUG_RUN_MAIN') != 'true':
        # If we have an import path we can print it out now which can help
        # people understand what's being served.  If we do not have an
        # import path because the app was loaded through a callback then
        # we won't print anything.
        if info.app_import_path is not None:
            print(' * Serving Flask app "%s"' % info.app_import_path)
        if debug is not None:
            print(' * Forcing debug mode %s' % (debug and 'on' or 'off'))

    run_simple(
        host, port, app, use_reloader=reload, use_debugger=debugger,
        threaded=with_threads, extra_files=template_paths(),
        request_handler=CustomLoggingWSGIRequestHandler
    )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import pytest
from biweeklybudget.settings import PAY_PERIOD_START_DATE
from biweeklybudget.utils import dtnow


@pytest.mark.acceptance
class Test00AcceptanceSaneSettings(object):

    def test_pay_period_start_date(self):
        assert PAY_PERIOD_START_DATE < dtnow().date()

    def test_settings_module(self):
        assert os.environ[
            'SETTINGS_MODULE'] == 'biweeklybudget.tests.fixtures.test_settings'
